import boto3
import json
import logging
import os
import pandas
import requests
import uuid

from io import StringIO
from prsn_athena.logger import get_logger
from typing import Any, Dict, List


class AthenaConnection:
    def __init__(
        self,
        database: str,
        workgroup: str,
        output_location: str = None,
        boto3_client=None,
        logger: logging.Logger = None
    ):
        try:
            self._database = database
            self._workgroup = workgroup
            self._output_location = output_location
            self._logger = logger if logger else get_logger('AthenaConnection')
            self._boto3_client = boto3_client if boto3_client else boto3.client('athena')

            if not self._output_location:
                resp = self._boto3_client.get_work_group(
                    WorkGroup=self._workgroup
                )
                wg_cfg = resp.get('WorkGroup', {}).get('Configuration', {})
                output_location = wg_cfg.get('ResultConfiguration', {}).get('OutputLocation')
                if output_location is None:
                    raise Exception(f'No output location found for workgroup {self._workgroup}')
                self._output_location = output_location
        except Exception as e:
            raise Exception(f'Failed to initialize an AthenaConnection with exception: {e}')

    @property
    def database(self) -> str:
        return self._database

    @property
    def workgroup(self) -> str:
        return self._workgroup

    @property
    def output_location(self) -> str:
        return self._output_location

    def execute_query(self, sql: str) -> Dict[str, Any]:
        self._logger.debug(f'Executing sql:\n{sql}')
        try:
            resp = self._boto3_client.start_query_execution(
                QueryString=sql,
                QueryExecutionContext={
                    'Database': self._database
                },
                ResultConfiguration={
                    'OutputLocation': self._output_location
                },
                WorkGroup=self._workgroup
            )

            result = {
                'query_execution_id': resp['QueryExecutionId'],
                'query_execution_status': 'QUEUED',
                'query_execution_start': None,
                'query_execution_end': None,
                'query_results': None
            }
        except Exception as e:
            self._logger.error(f'Error while executing query: {e}')
            raise e

        return result

    def get_query_results(self, query_execution_id: str):
        self._logger.debug(f'Retrieving results for query {query_execution_id}')
        try:
            resp = self._boto3_client.get_query_execution(
                QueryExecutionId=query_execution_id)
            query_execution = resp.get('QueryExecution', {})
            status = query_execution['Status']['State']
            start_time = query_execution['Status'].get('SubmissionDateTime')
            end_time = query_execution['Status'].get('CompletionDateTime')
            result = {
                'query_execution_id': query_execution_id,
                'query_execution_status': status,
                'query_execution_start': None,
                'query_execution_end': None,
                'query_results': None
            }
            if start_time:
                result['query_execution_start'] = start_time.strftime("%Y-%m-%dT%H:%M:%SZ")
            if end_time:
                result['query_execution_end'] = end_time.strftime("%Y-%m-%dT%H:%M:%SZ")
            if status in ['SUCCEEDED']:
                result['query_results'] = self.get_signed_url(
                    query_execution_id=query_execution_id,
                    location=query_execution['ResultConfiguration']['OutputLocation']
                )
            elif status in ['QUEUED', 'RUNNING', 'FAILED', 'CANCELLED']:
                pass
            else:
                raise Exception(f'Unrecognized athena query status: {status}')
            return result
        except Exception as e:
            self._logger.error(f'Error while getting query execution {query_execution_id}: {e}')
            raise e

    def get_signed_url(self, query_execution_id: str, location: str) -> str:
        self._logger.debug(f'Getting a signed URL for query_execution_id: {query_execution_id}')
        try:
            resp = boto3.client('s3').generate_presigned_url(
                'get_object',
                Params={
                    'Bucket': trim_prefix(location, 's3://').split('/', 1)[0],
                    'Key': trim_prefix(location, 's3://').split('/', 1)[1]
                },
                ExpiresIn=900
            )
        except Exception as e:
            self._logger.error(f'Failed to retrieved signed url with error: {e}')
            raise e
        return resp

    def download_results(self, s3_url: str, filepath: str):
        self._logger.debug(f'Downloading file from signed url {s3_url}')
        try:
            with requests.get(s3_url, stream=True) as resp:
                resp.raise_for_status()
                with open(filepath, 'wb') as f:
                    for chunk in resp.iter_content(chunk_size=8192):
                        f.write(chunk)
        except Exception as e:
            self._logger.error(f'Failed to download S3 file with exception: {e}')
            raise e

    def download_results_as_dict(self, s3_url: str, filepath: str = None) -> List[Any]:
        self._logger.debug(f'Downloading results as dict')
        try:
            filepath = filepath if filepath else f'/tmp/{uuid.uuid4().hex}.csv'
            self.download_results(s3_url, filepath)
            with open(filepath, 'r') as f:
                df = pandas.read_csv(StringIO(f.read()), dtype=str)
            os.remove(filepath)
            return df.to_dict('records')
        except Exception as e:
            self._logger.error(f'Failed downloading results as dict with exception: {e}')
            raise e

    def to_dict(self) -> Dict[str, str]:
        return {
            'database': self._database,
            'workgroup': self._workgroup,
            'output_location': self._output_location
        }

    def __repr__(self) -> str:
        return json.dumps(self.to_dict())


def trim_prefix(s: str, prefix: str):
    if s.startswith(prefix):
        return s[len(prefix):]
    return s



import logging
import sys


def get_logger(name: str, log_level: str = 'info') -> logging.Logger:
    """Return a logger with the given name and logging level set"""
    handler = logging.StreamHandler(sys.stdout)
    log = logging.getLogger(name)
    log_level = log_level.lower()
    if log_level == 'debug':
        log.setLevel(logging.DEBUG)
        handler.setLevel(logging.DEBUG)
    elif log_level == 'warning':
        log.setLevel(logging.WARNING)
        handler.setLevel(logging.WARNING)
    elif log_level == 'error':
        log.setLevel(logging.ERROR)
        handler.setLevel(logging.ERROR)
    elif log_level == 'critical':
        log.setLevel(logging.CRITICAL)
        handler.setLevel(logging.CRITICAL)
    else:
        log.setLevel(logging.INFO)
        handler.setLevel(logging.INFO)

    formatter = logging.Formatter(
        '%(asctime)s - %(funcName)s - %(lineno)i [%(levelname)s] %(message)s')
    handler.setFormatter(formatter)
    log.addHandler(handler)
    log.propagate = False
    return log



import json

from typing import Any, List


class QueryBuilder:
    def __init__(self, table_name: str, projection_attributes: List[str] = []):
        self._table_name = table_name
        self._projection_attributes = projection_attributes
        self._clauses = []

    @property
    def table_name(self):
        return self._table_name

    @property
    def projection_attributes(self) -> List[str]:
        return self._projection_attributes

    @property
    def clauses(self) -> List[str]:
        return self._clauses

    def add_string_clause(
        self,
        attribute: str,
        values: List[str]
    ) -> str:
        '''
        >>> qb = QueryBuilder('table_name')
        >>> qb.add_string_clause('t_environment', []) is None
        True

        >>> qb.add_string_clause('t_environment', ['PRD'])
        "t_environment = 'PRD'"

        >>> qb.add_string_clause('t_dcl', [1, 2])
        "t_dcl IN ('1','2')"

        >>> len(qb.clauses)
        2
        '''
        clause = None
        if attribute and values:
            if len(values) == 1:
                clause = f"{attribute} = '{values[0]}'"
            else:
                clause = f'{attribute} IN ('
                for val in sorted(values):
                    clause = f"{clause}'{val}',"
                clause = clause[:-1]  # remove hanging comma
                clause = f'{clause})'
            self._clauses.append(clause)
        return clause

    def add_numeric_clause(
        self,
        attribute: str,
        op: str,
        val: float,
        attribute_is_string: bool = False,
        places: int = None,
        precision: int = None
    ) -> str:
        '''
        >>> qb = QueryBuilder('table_name')
        >>> qb.add_numeric_clause('count', '=', None) is None
        True

        >>> qb.add_numeric_clause('count', '!=', 3.0)
        Traceback (most recent call last):
        AssertionError: Unrecognized numeric operator: !=

        >>> qb.add_numeric_clause('count', '=', 3.0)
        'count = 3.0'

        >>> qb.add_numeric_clause('count', '<', 4.0, attribute_is_string=True)
        'TRY_CAST(count AS DECIMAL) < 4.0'

        >>> qb.add_numeric_clause('count', '>', 1.0, attribute_is_string=True, places=4, precision=2)
        'TRY_CAST(count AS DECIMAL(4, 2)) > 1.0'

        >>> len(qb.clauses)
        3
        '''
        clause = None
        assert op in ['=', '<>', '<', '>', '<=', '>='], f'Unrecognized numeric operator: {op}'
        if attribute and op and val:
            if attribute_is_string and places and precision:
                clause = f'TRY_CAST({attribute} AS DECIMAL({places}, {precision})) {op} {val}'
            elif attribute_is_string:
                clause = f'TRY_CAST({attribute} AS DECIMAL) {op} {val}'
            else:
                clause = f'{attribute} {op} {val}'
            self._clauses.append(clause)
        return clause

    def add_numeric_list_clause(
        self,
        attribute: str,
        values: List[Any],
        attribute_is_string: bool = False,
        places: int = None,
        precision: int = None
    ) -> str:
        '''
        >>> qb = QueryBuilder('table_name')
        >>> qb.add_numeric_list_clause('count', []) is None
        True

        >>> qb.add_numeric_list_clause('count', [1])
        'count = 1'

        >>> qb.add_numeric_list_clause('recount', [2, 3])
        'recount IN (2,3)'

        >>> qb.add_numeric_list_clause('count', [4], attribute_is_string=True)
        'TRY_CAST(count AS DECIMAL) = 4'

        >>> qb.add_numeric_list_clause('count', [5], attribute_is_string=True, places=5, precision=3)
        'TRY_CAST(count AS DECIMAL(5, 3)) = 5'

        >>> len(qb.clauses)
        4
        '''
        clause = None
        if attribute and values:
            if attribute_is_string and places and precision:
                clause = f'TRY_CAST({attribute} AS DECIMAL({places}, {precision}))'
            elif attribute_is_string:
                clause = f'TRY_CAST({attribute} AS DECIMAL)'
            else:
                clause = f'{attribute}'

            if len(values) == 1:
                clause = f"{clause} = {values[0]}"
            else:
                clause = f'{clause} IN ('
                for val in sorted(values):
                    clause = f"{clause}{val},"
                clause = clause[:-1]  # remove hanging comma
                clause = f'{clause})'
            self._clauses.append(clause)
        return clause

    def add_date_clause(
        self,
        attribute: str,
        op: str,
        val: str,
        attribute_is_string: bool = False
    ) -> str:
        '''
        >>> qb = QueryBuilder('table_name')
        >>> qb.add_date_clause('report_date', '=', None) is None
        True

        >>> qb.add_date_clause('report_date', '!=', '2021-04-25')
        Traceback (most recent call last):
        AssertionError: Unrecognized date operator: !=

        >>> qb.add_date_clause('report_date', '>', '2021-04-25')
        "report_date > DATE('2021-04-25')"

        >>> qb.add_date_clause('found_date', '<', '2021-06-15', attribute_is_string=True)
        "TRY_CAST(SUBSTRING(found_date, 1, 10) AS DATE) < DATE('2021-06-15')"

        >>> len(qb.clauses)
        2
        '''
        clause = None
        assert op in ['=', '<>', '<', '>', '<=', '>='], f'Unrecognized date operator: {op}'
        if attribute and op and val:
            if attribute_is_string:
                clause = f"TRY_CAST(SUBSTRING({attribute}, 1, 10) AS DATE) {op} DATE('{val}')"
            else:
                clause = f"{attribute} {op} DATE('{val}')"
            self._clauses.append(clause)
        return clause

    def render_sql(self) -> str:
        '''
        >>> qb = QueryBuilder('table_name', ['account', 'region'])
        >>> qb.render_sql()
        'SELECT account, region FROM table_name;'

        >>> qb = QueryBuilder('table_name')
        >>> qb.render_sql()
        'SELECT * FROM table_name;'

        >>> _ = qb.add_string_clause('user', ['nelson.mandela@sa.com'])
        >>> qb.render_sql()
        "SELECT * FROM table_name WHERE user = 'nelson.mandela@sa.com';"

        >>> _ = qb.add_numeric_clause('count', '=', 10, attribute_is_string=True)
        >>> qb.render_sql()
        "SELECT * FROM table_name WHERE TRY_CAST(count AS DECIMAL) = 10 AND user = 'nelson.mandela@sa.com';"
        '''
        projection_attributes = '*'
        if self._projection_attributes:
            sep = ', '
            projection_attributes = sep.join(self._projection_attributes)

        sql = f'SELECT {projection_attributes} FROM {self._table_name}'
        if self._clauses:
            sql += ' WHERE '
            sep = ' AND '
            sql += sep.join(sorted(self._clauses))
        return f'{sql};'

    def __repr__(self) -> str:
        return json.dumps({
            'table_name': self._table_name,
            'projection_attributes': self._projection_attributes,
            'clauses': self._clauses
        })


if __name__ == '__main__':
    import doctest
    doctest.testmod()



# coding: utf-8

"""
    Pearson Athena

     Module containing a SQL query builder for Athena and an Athena connection that simplifies interactions with AWS Athena.   # noqa: E501
"""


from setuptools import setup, find_packages  # noqa: H301

NAME = "prsn-athena"
VERSION = "0.2.0"
# To install the library, run the following
#
# python setup.py install
#
# prerequisite: setuptools
# http://pypi.python.org/pypi/setuptools

REQUIRES = ["boto3 >= 1.14.63", "pandas >= 1.0.5", "requests >= 2.23.0"]

setup(
    name=NAME,
    version=VERSION,
    description="Pearson Athena",
    author="Pearson Cloud Management",
    author_email="io-pearson-cloud-management@pearson.com",
    url="",
    keywords=["Pearson", "PCM", "Athena"],
    install_requires=REQUIRES,
    packages=find_packages(),
    include_package_data=True,
    long_description="""\
     Module containing a SQL query builder for Athena and an Athena connection that simplifies interactions with AWS Athena.   # noqa: E501
    """
)
